;;; GNU Guix --- Functional package management for GNU
;;; Copyright © 2016 Roel Janssen <roel@gnu.org>
;;;
;;; This file is part of GNU Guix.
;;;
;;; GNU Guix is free software; you can redistribute it and/or modify it
;;; under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or (at
;;; your option) any later version.
;;;
;;; GNU Guix is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Guix.  If not, see <http://www.gnu.org/licenses/>.

(define-module (guix workflows)
  #:use-module (ice-9 match)
  #:use-module (srfi srfi-9 gnu)
  #:use-module (srfi srfi-35)
  #:use-module (guix utils)
  #:use-module (guix build utils)
  #:use-module (guix monads)
  #:use-module (guix gexp)
  #:use-module (guix store)
  #:use-module (guix records)
  #:use-module (guix packages)
  #:use-module (guix workflows execution-order)
  #:use-module (guix workflows execution-helpers)
  #:use-module (guix workflows utils)
  #:use-module (gnu packages)
  #:use-module (gnu packages guile)
  #:use-module (gnu packages bash)

  #:re-export (sequential-execution-order)
  #:re-export (parallel-step-execution-order)
  #:re-export (succesful-execution?)
  #:re-export (source->string)

  #:export (workflow
            workflow?
            workflow-name
            workflow-full-name
            workflow-version
            workflow-input
            workflow-output
            workflow-processes
            workflow-restrictions
            workflow-arguments
            workflow-execution
            workflow-synopsis
            workflow-description

            process
            process?
            process-name
            process-full-name
            process-version
            process-environment
            process-input
            process-output
            process-procedure
            process-synopsis
            process-description
            process-output-of
            execute-process

            script
            script?
            script-interpreter
            script-source

            execute-workflow
            workflow->dot
            workflow-order->dot))

;;; Commentary:
;;;
;;; This module provides a mechanism to define workflows, which are essentially
;;; a sequential execution of programs provided by Guix packages, with variable
;;; input and output data.  The sequence can be partly executed in parallel.
;;;
;;; Workflows can be used to build automated test suites with real data, without
;;; cluttering the source repository with this data.  This module was originally
;;; designed to accommodate the creation of bioinformatics pipelines.
;;;
;;; Each program invocation is considered a process.  Dependencies between
;;; processes exist when process A uses the output of process B as its input.
;;;

;; Workflow record type:
;; ----------------------------------------------------------------------------
;; This record type captures the processes, input data and expected output data
;; of a workflow.  The idea behind a workflow is that it can be executed like a
;; regular program.
;;
(define-record-type* <workflow>
  workflow make-workflow
  workflow?

  ;; Basic information about the workflow
  (name workflow-name)
  (version workflow-version)
  (synopsis workflow-synopsis)
  (description workflow-description)

  ;; The input and output of a workfow will be passed to each process as
  ;; the base input and output directories.
  (input workflow-input)
  (output workflow-output)

  ;; Processes are functions taking two parameters (input and output) that
  ;; return a <process> record type.
  (processes workflow-processes)

  ;; Processes can depend on each other.  By defining dependency pairs
  ;; in the form (A B) where A must be executed after B.
  ;;
  ;; When automatic-restrictions is set to #t, we do not need to specify
  ;; workflow-restrictions manually.  The dependencies will be computed
  ;; automatically.
  ;;
  ;; Automatically computing the dependencies may be too simplistic for
  ;; your workflow.
  (restrictions workflow-restrictions (default '()))
  (automatic-restrictions workflow-automatic-restrictions
                          (default #f))
  
  ;; Arguments are literally command-line arguments that can be passed
  ;; when executing a specific workflow.  This allows users to turn features
  ;; off and on, and pass an input and output directory.
  ;;
  ;; The arguments are passed as a list to the workflow record.
  (arguments workflow-arguments
             (default #f))

  ;; When a workflow requires additional code to execute, it can be
  ;; specified in the following field.
  (execution workflow-execution
             (default #f)))

(define (workflow-full-name process port)
  "Writes the name and version as a single string of PROCESS to PORT."
  (simple-format port "~a-~a" (workflow-name process)
                              (workflow-version process)))

(define (print-workflow workflow port)
  "Write a decent human-representation of a workflow of WORKFLOW to PORT."
  (simple-format port "#<workflow ~a>" (workflow-full-name workflow #f)))

(set-record-type-printer! <workflow> print-workflow)

;; Process record type:
;; ----------------------------------------------------------------------------
;;
;; A process is a stand-alone unit doing some work.  This is typically a single
;; program, configured through command-line options, given a certain input,
;; producing a certain output.
;;
(define-record-type* <process>
  process make-process
  process?

  ;; Basic information about the process
  (name process-name)
  (version process-version)
  (synopsis process-synopsis)
  (description process-description)

  (input process-input)
  (output process-output)

  ;; The environment is a list of package variables that will be installed
  ;; in an isolated environment, in which the procedure will be executed.
  (environment process-environment)

  ;; The imperative procedure which most probably executes a program
  ;; on the operating system.
  (procedure process-procedure))

(define (process-full-name process)
  "Writes the name and version as a single string of PROCESS to PORT."
  (string-append (process-name process) "-" (process-version process)))

;; (define (process-output-of process)
;;   "Returns the expected output filename(s) of a process."
;;   (data-argument-resource (process-output process)))

;; (define (process-procedure-source process)
;;   "Returns the source code for the procedure within PROCESS."
;;   (procedure-property (process-procedure process) 'source))

;; FIXME: Currently, the output parameter can only be a directory.
;; FIXME: The shell header is not flexible enough.  For Guile we need three
;;        lines.  It would be preferable to use the Guix path for pointing
;;        to an interpreter.
(define (execute-process process)
  "Calls the execution command for PROCESS when passed the INPUT and OUTPUT
parameters."
  (unless (eq? (process-procedure process) #f)
    (let* ((name (process-name process))
           (template (string-append "/tmp/guix-process-" name "-XXXXXX"))
           (port (mkstemp! template))
           (filename (port-filename port))
           (proc (process-procedure process))
           (output (process-output process))
           (interpreter (script-interpreter proc))
           (guile-bin ((package-file guile-2.0 "bin/guile") (open-connection))))
      (format port "~a~%~a~%"
              (generate-script-header interpreter)
              ;; Leave the creation of an output directory up to the process.
              ;(interpreter-specific-mkdir interpreter output)
              (source->string (script-source proc)))
      (close port)
      (format #t "Executing: ~a ~a ~a ~a~%"
              guile-bin "--no-auto-compile" "-c"
              (format #f "~s" (source->string (script-source proc))))
      (eval (script-source proc) (interaction-environment))
      ;;(eqv? 0 (status:exit-val (system* guile-bin "-c"
      ;;                                  (source->string (script-source proc)))))
      ;;(eqv? 0 (status:exit-val (system* (package-file guile-2.0 "bin/guile") filename)))
      )))

(define (print-process process port)
  "Write a decent human-readable representation of a PROCESS to PORT."
  (simple-format port "#<process ~a>" (process-full-name process)))

(set-record-type-printer! <process> print-process)

(define-record-type* <script>
  script make-script
  script?

  (interpreter script-interpreter)
  (source script-source))

(define (print-script script port)
  "Write a decent human-readable representation of SCRIPT to PORT."
  (simple-format port "#<script ~a>" (script-interpreter script)))

(set-record-type-printer! <script> print-script)

;;
;; Functionality operating on the record types defined above.
;;
(define (execute-workflow workflow)
  "Returns a list of commands to run for WORKFLOW."
  (let ((order (sequential-execution-order
                (workflow-processes workflow)
                (workflow-restrictions workflow))))
    (if (null? order)
      (display "Error: Circular dependency found in the workflow restrictions.")
      (begin
        (format #t "Progress:~%")
        (map (lambda (process)
               (let* ((function (eval process (interaction-environment)))
                      (proc (function (workflow-input workflow)
                                      (workflow-output workflow))))
                 (format #t "  Executing ~a~%" (process-full-name proc))
                 (execute-process (proc (workflow-input workflow)
                                        (workflow-output workflow)))))
             order)
        (newline)))))

(define take-color (color-scheme-stepper %light-color-scheme))

(define (workflow-dot-prettify-node process)
  "Returns a string of prettified node names for a Graphviz graph."
  (let* ((function (eval process (interaction-environment)))
         (proc (function "" ""))
         (pretty-name (string-map (lambda (x)
                                    (if (eq? x #\-) #\  x))
                                  (process-name proc))))
    (format #f " ~s [shape=box,style=\"rounded,filled\",fillcolor=~s,label=<<FONT POINT-SIZE=\"14\"><U>~a</U></FONT><BR/><FONT POINT-SIZE=\"12\">~a<BR/><BR/>Uses: ~{~a~^, ~}.</FONT>>];~%"
            (process-full-name proc)
            (take-color)
            (string-upcase pretty-name)
            (process-synopsis proc)
            (map (lambda (package)
                   (if (> (length package) 0)
                       (package-full-name (list-ref package 1))
                       ""))
                 (process-environment proc)))))

(define (workflow->dot workflow)
  "Returns the workflow's processes formatted in Graphviz's Dot language as a
directed acyclic graph."
  (format #f "digraph G {~%  graph [label=<~a<BR/><BR/>>,labelloc=t,labeljust=center,fontsize=24];~%~{~a~}~%~{~a~}}"
          (string-upcase (string-map (lambda (x)
                                       (if (eq? x #\-) #\  x))
                                     (workflow-name workflow)))
          (map workflow-dot-prettify-node (workflow-processes workflow))
          (map (lambda (pair)
                 (let ((first (list-ref pair 0))
                       (second (list-ref pair 1))
                       (env (interaction-environment)))
                   (format #f "  ~s -> ~s~%"
                           ;; Reversed order here because the flow is exactly
                           ;; the opposite of the dependency chain.
                           (process-full-name ((eval second env) "" ""))
                           (process-full-name ((eval first env) "" "")))))
               (workflow-restrictions workflow))))

(define (workflow-sequential-order->dot workflow)
  (format #f "digraph G {~% ~{~a~}~%~a~%}"
          (map (lambda (process)
                 (let* ((function (eval process (interaction-environment)))
                        (proc (function "" ""))
                        (pretty-name (string-map (lambda (x)
                                                   (if (eq? x #\-) #\  x))
                                                 (process-name proc))))
                   (format #f " ~s [shape=box,style=\"rounded,filled\",fillcolor=~s,label=<~a>]~%"
                           (process-full-name proc)
                           (take-color)
                           (string-upcase pretty-name))))
               (workflow-processes workflow))
          (format #f "~{  ~s ~^->~%~}"
                  (reverse (map (lambda (process)
                                  (let* ((function (eval process (interaction-environment)))
                                         (proc (function "" "")))
                                    (process-full-name proc)))
                                (sequential-execution-order
                                 (workflow-processes workflow)
                                 (workflow-restrictions workflow)))))))
